@@ -96,20 +96,35 @@ document 6
 ...     print result['name']
 document 5
 
+Docs can also boost any particular key. This lets us use Solr's boost on a field.
+
+>>> docs = [
+...     {'id': 'testdoc.7', 'order_i': '7', 'name': 'document 7', 'text': u'eight', 'author': 'seven'},
+...     {'id': 'testdoc.8', 'order_i': '8', 'name': 'document 8', 'text': u'seven', 'author': 'eight'},
+... ]
+
+>>> conn.add(docs, boost={'author': '2.0',})
+>>> results = conn.search('seven author:seven')
+>>> for result in results:
+...     print result['name']
+document 7
+document 8
+
+>>> results = conn.search('eight author:eight')
+>>> for result in results:
+...     print result['name']
+document 8
+document 7
+
 """
 
 # TODO: unicode support is pretty sloppy. define it better.
 
-from urllib import urlencode
-from urlparse import urlsplit, urlunsplit
+from datetime import datetime
 import re
-
-try:
-    # Use Django's implementation if available because it can handle dates before 1900.
-    from django.utils.datetime_safe import datetime, date
-except ImportError:
-    # Fall back to the default
-    from datetime import datetime, date
+import urllib
+import urllib2
+from urlparse import urlsplit, urlunsplit
 
 try:
     # for python 2.5
@@ -150,10 +165,10 @@ except NameError:
 
 __author__ = 'Joseph Kocherhans, Jacob Kaplan-Moss, Daniel Lindsley'
 __all__ = ['Solr']
-__version__ = (2, 0, 11)
+__version__ = (2, 0, 12)
 
 def get_version():
-    return "%s.%s.%s" % __version__
+    return "%s.%s.%s" % __version__[0:3]
 
 DATETIME_REGEX = re.compile('^(?P<year>\d{4})-(?P<month>\d{2})-(?P<day>\d{2})T(?P<hour>\d{2}):(?P<minute>\d{2}):(?P<second>\d{2})(\.\d+)?Z$')
 
@@ -161,12 +176,13 @@ class SolrError(Exception):
     pass
 
 class Results(object):
-    def __init__(self, docs, hits, highlighting=None, facets=None, spellcheck=None):
+    def __init__(self, docs, hits, highlighting=None, facets=None, spellcheck=None, stats=None):
         self.docs = docs
         self.hits = hits
         self.highlighting = highlighting or {}
         self.facets = facets or {}
         self.spellcheck = spellcheck or {}
+        self.stats = stats or {}
 
     def __len__(self):
         return len(self.docs)
@@ -192,7 +208,13 @@ class Solr(object):
     def _send_request(self, method, path, body=None, headers=None):
         if TIMEOUTS_AVAILABLE:
             http = Http(timeout=self.timeout)
-            headers, response = http.request(self.base_url + path, method=method, body=body, headers=headers)
+            url = self.base_url + path
+            
+            try:
+                headers, response = http.request(url, method=method, body=body, headers=headers)
+            except AttributeError:
+                # For httplib2.
+                raise SolrError("Failed to connect to server at '%s'. Are you sure '%s' is correct? Checking it in a browser might help..." % (url, self.base_url))
             
             if int(headers['status']) != 200:
                 raise SolrError(self._extract_error(headers, response))
@@ -219,25 +241,25 @@ class Solr(object):
         
         if len(params['q']) < 1024:
             # Typical case.
-            path = '%s/select/?%s' % (self.path, urlencode(params, True))
+            path = '%s/select/?%s' % (self.path, urllib.urlencode(params, True))
             return self._send_request('GET', path)
         else:
             # Handles very long queries by submitting as a POST.
-            path = '%s/select/?%s' % (self.path,)
+            path = '%s/select/' % (self.path,)
             headers = {
-                'Content-type': 'application/x-www-form-urlencoded',
+                'Content-type': 'application/x-www-form-urlencoded; charset=utf-8',
             }
-            body = urlencode(params, False)
+            body = urllib.urlencode(params, False)
             return self._send_request('POST', path, body=body, headers=headers)
     
     def _mlt(self, params):
         # encode the query as utf-8 so urlencode can handle it
         params['q'] = params['q'].encode('utf-8')
         params['wt'] = 'json' # specify json encoding of results
-        path = '%s/mlt/?%s' % (self.path, urlencode(params, True))
+        path = '%s/mlt/?%s' % (self.path, urllib.urlencode(params, True))
         return self._send_request('GET', path)
 
-    def _update(self, message, clean_ctrl_chars=True):
+    def _update(self, message, clean_ctrl_chars=True, commit=True):
         """
         Posts the given xml message to http://<host>:<port>/solr/update and
         returns the result.
@@ -249,22 +271,97 @@ class Solr(object):
         """
         path = '%s/update/' % self.path
         
+        # Per http://wiki.apache.org/solr/UpdateXmlMessages, we can append a
+        # ``commit=true`` to the URL and have the commit happen without a
+        # second request.
+        if commit:
+            path += '?commit=true'
+        
         # Clean the message of ctrl characters.
         if clean_ctrl_chars:
             message = sanitize(message)
         
-        return self._send_request('POST', path, message, {'Content-type': 'text/xml'})
-
+        return self._send_request('POST', path, message, {'Content-type': 'text/xml; charset=utf-8'})
+    
     def _extract_error(self, headers, response):
         """
-        Extract the actual error message from a solr response. Unfortunately,
-        this means scraping the html.
+        Extract the actual error message from a solr response.
         """
-        jetty_br = '<br/>                                                \n'
-        return "[Reason: %s]\n%s" % (headers.get('reason'), response.replace(jetty_br, ''))
-
+        reason = headers.get('reason', None)
+        full_html = None
+        
+        if reason is None:
+            reason, full_html = self._scrape_response(headers, response)
+        
+        msg = "[Reason: %s]" % reason
+        
+        if reason is None:
+            msg += "\n%s" % full_html
+        
+        return msg
+    
+    def _scrape_response(self, headers, response):
+        """
+        Scrape the html response.
+        """
+        # identify the responding server
+        server_type = None
+        server_string = headers.get('server', None)
+        
+        if 'jetty' in server_string.lower():
+            server_type = 'jetty'
+        
+        if 'coyote' in server_string.lower():
+            # TODO: During the pysolr 3 effort, make this no longer a
+            #       conditional and consider using ``lxml.html`` instead.
+            from BeautifulSoup import BeautifulSoup
+            server_type = 'tomcat'
+        
+        reason = None
+        full_html = ''
+        dom_tree = None
+        
+        if server_type == 'tomcat':
+            # Tomcat doesn't produce a valid XML response
+            soup = BeautifulSoup(response)
+            body_node = soup.find('body')
+            p_nodes = body_node.findAll('p')
+            
+            for p_node in p_nodes:
+                children = p_node.findChildren()
+                
+                if len(children) >= 2 and 'message' in children[0].renderContents().lower():
+                    reason = children[1].renderContents()
+            
+            if reason is None:
+                full_html = soup.prettify()
+        else:
+            # Let's assume others do produce a valid XML response
+            try:
+                dom_tree = ET.fromstring(response)
+                reason_node = None
+                
+                # html page might be different for every server
+                if server_type == 'jetty':
+                    reason_node = dom_tree.find('body/pre')
+                
+                if reason_node is not None:
+                    reason = reason_node.text
+                
+                if reason is None:
+                    full_html = ET.tostring(dom_tree)
+            except SyntaxError, e:
+                full_html = "%s" % response
+        
+        full_html = full_html.replace('\n', '')
+        full_html = full_html.replace('\r', '')
+        full_html = full_html.replace('<br/>', '')
+        full_html = full_html.replace('<br />', '')
+        full_html = full_html.strip()
+        return reason, full_html
+    
     # Conversion #############################################################
-
+    
     def _from_python(self, value):
         """
         Converts python values to a form suitable for insertion into the xml
@@ -272,9 +369,9 @@ class Solr(object):
         """
         if hasattr(value, 'strftime'):
             if hasattr(value, 'hour'):
-                value = value.strftime('%Y-%m-%dT%H:%M:%SZ')
+                value = "%sZ" % value.isoformat()
             else:
-                value = value.strftime('%Y-%m-%dT00:00:00Z')
+                value = "%sT00:00:00Z" % value.isoformat()
         elif isinstance(value, bool):
             if value:
                 value = 'true'
@@ -324,9 +421,19 @@ class Solr(object):
             pass
         
         return value
-
+    
+    def _is_null_value(self, value):
+        """
+        Check if a given value is ``null``.
+        
+        Criteria for this is based on values that shouldn't be included
+        in the Solr ``add`` request at all.
+        """
+        # TODO: This should probably be removed when solved in core Solr level?
+        return (value is None) or (isinstance(value, basestring) and len(value) == 0)
+    
     # API Methods ############################################################
-
+    
     def search(self, q, **kwargs):
         """Performs a search and returns the results."""
         params = {'q': q}
@@ -346,6 +453,9 @@ class Solr(object):
         if result.get('spellcheck'):
             result_kwargs['spellcheck'] = result['spellcheck']
         
+        if result.get('stats'):
+            result_kwargs['stats'] = result['stats']
+        
         return Results(result['response']['docs'], result['response']['numFound'], **result_kwargs)
     
     def more_like_this(self, q, mltfl, **kwargs):
@@ -370,35 +480,51 @@ class Solr(object):
             }
         
         return Results(result['response']['docs'], result['response']['numFound'])
-
-    def add(self, docs, commit=True):
-        """Adds or updates documents. For now, docs is a list of dictionaies
+    
+    def add(self, docs, commit=True, boost=None):
+        """Adds or updates documents. For now, docs is a list of dictionaries
         where each key is the field name and each value is the value to index.
         """
         message = ET.Element('add')
+        
         for doc in docs:
             d = ET.Element('doc')
+            
             for key, value in doc.items():
-                # handle lists, tuples, and other iterabes
+                if key == 'boost':
+                    d.set('boost', str(value))
+                    continue
+                
+                # handle lists, tuples, and other iterables
                 if hasattr(value, '__iter__'):
                     for v in value:
-                        f = ET.Element('field', name=key)
+                        if self._is_null_value(value):
+                            continue
+                        
+                        if boost and v in boost:
+                            f = ET.Element('field', name=key, boost=boost[v])
+                        else:
+                            f = ET.Element('field', name=key)
+                        
                         f.text = self._from_python(v)
                         d.append(f)
                 # handle strings and unicode
                 else:
-                    f = ET.Element('field', name=key)
+                    if self._is_null_value(value):
+                        continue
+                    
+                    if boost and key in boost:
+                        f = ET.Element('field', name=key, boost=boost[key])
+                    else:
+                        f = ET.Element('field', name=key)
                     f.text = self._from_python(value)
                     d.append(f)
+            
             message.append(d)
-        m = ET.tostring(message)
-        response = self._update(m)
-        # TODO: Supposedly, we can put a <commit /> element in the same post body
-        # as the add element. That isn't working for some reason, and it would save us
-        # an extra trip to the server. This works for now.
-        if commit:
-            self.commit()
-
+        
+        m = ET.tostring(message, 'utf-8')
+        response = self._update(m, commit=commit)
+    
     def delete(self, id=None, q=None, commit=True, fromPending=True, fromCommitted=True):
         """Deletes documents."""
         if id is None and q is None:
@@ -409,20 +535,119 @@ class Solr(object):
             m = '<delete><id>%s</id></delete>' % id
         elif q is not None:
             m = '<delete><query>%s</query></delete>' % q
-        response = self._update(m)
-        # TODO: Supposedly, we can put a <commit /> element in the same post body
-        # as the delete element. That isn't working for some reason, and it would save us
-        # an extra trip to the server. This works for now.
-        if commit:
-            self.commit()
-
+        
+        response = self._update(m, commit=commit)
+    
     def commit(self):
         response = self._update('<commit />')
-
+    
     def optimize(self):
         response = self._update('<optimize />')
 
 
+class SolrCoreAdmin(object):
+    """
+    Handles core admin operations: see http://wiki.apache.org/solr/CoreAdmin
+    
+    Operations offered by Solr are:
+       1. STATUS
+       2. CREATE
+       3. RELOAD
+       4. RENAME
+       5. ALIAS
+       6. SWAP
+       7. UNLOAD
+       8. LOAD (not currently implemented)
+    """
+    def __init__(self, url, *args, **kwargs):
+        super(SolrCoreAdmin, self).__init__(*args, **kwargs)
+        self.url = url
+    
+    def _get_url(self, url, params={}, headers={}):
+        request = urllib2.Request(url, data=urllib.urlencode(params), headers=headers)
+        # Let ``socket.error``, ``urllib2.HTTPError`` and ``urllib2.URLError``
+        # propagate up the stack.
+        response = urllib2.urlopen(request)
+        return response.read()
+    
+    def status(self, core=None):
+        """http://wiki.apache.org/solr/CoreAdmin#head-9be76f5a459882c5c093a7a1456e98bea7723953"""
+        params = {
+            'action': 'STATUS',
+        }
+        
+        if core is not None:
+            params.update(core=core)
+        
+        return self._get_url(self.url, params=params)
+    
+    def create(self, name, instance_dir=None, config='solrcofig.xml', schema='schema.xml'):
+        """http://wiki.apache.org/solr/CoreAdmin#head-7ca1b98a9df8b8ca0dcfbfc49940ed5ac98c4a08"""
+        params = {
+            'action': 'STATUS',
+            'name': name,
+            'config': config,
+            'schema': schema,
+        }
+        
+        if instance_dir is None:
+            params.update(instanceDir=name)
+        else:
+            params.update(instanceDir=instance_dir)
+        
+        return self._get_url(self.url, params=params)
+    
+    def reload(self, core):
+        """http://wiki.apache.org/solr/CoreAdmin#head-3f125034c6a64611779442539812067b8b430930"""
+        params = {
+            'action': 'RELOAD',
+            'core': core,
+        }
+        return self._get_url(self.url, params=params)
+    
+    def rename(self, core, other):
+        """http://wiki.apache.org/solr/CoreAdmin#head-9473bee1abed39e8583ba45ef993bebb468e3afe"""
+        params = {
+            'action': 'RENAME',
+            'core': core,
+            'other': other,
+        }
+        return self._get_url(self.url, params=params)
+    
+    def alias(self, core, other):
+        """
+        http://wiki.apache.org/solr/CoreAdmin#head-8bf9004eaa4d86af23d2758aafb0d31e2e8fe0d2
+        
+        Experimental feature in Solr 1.3
+        """
+        params = {
+            'action': 'ALIAS',
+            'core': core,
+            'other': other,
+        }
+        return self._get_url(self.url, params=params)
+    
+    def swap(self, core, other):
+        """http://wiki.apache.org/solr/CoreAdmin#head-928b872300f1b66748c85cebb12a59bb574e501b"""
+        params = {
+            'action': 'SWAP',
+            'core': core,
+            'other': other,
+        }
+        return self._get_url(self.url, params=params)
+    
+    def unload(self, core):
+        """http://wiki.apache.org/solr/CoreAdmin#head-f5055a885932e2c25096a8856de840b06764d143"""
+        params = {
+            'action': 'UNLOAD',
+            'core': core,
+        }
+        return self._get_url(self.url, params=params)
+        
+    def load(self, core):
+        raise NotImplementedError('Solr 1.4 and below do not support this operation.')
+
+
 # Using two-tuples to preserve order.
 REPLACEMENTS = (
     # Nuke nasty control characters.
